/*
  Copyright 1997, 1998, 2002 Ronald S. Burkey.

  This file is part of RIP.

  RIP is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  RIP is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with RIP; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Filename:   	ripp.cpp
  Purpose:    	RIP-library access (not creation or updating) workhorse
         	functions for 32-bit compilers.
  Mods:       	12/20/97 RSB	Created by combining and pruning the files
                              	RIP.C, RIP_2.C, and RIPC.C used for the
                              	DOS version. 
		12/21/97 RSB	Changed to C++, though I don't really expect
               			to have any C++ code, because I couldn't use
                              	the rip97.h header otherwise.  Also, began
                              	actually interfacing the code to my user
                              	interface.
		12/22/97 RSB	Put in the most of the initial quick-search to
               			grab the list of candidate files in a db-search.
		12/24/97 RSB	Managed to get a list-box with RIP files in it
				from a query, though I can't uncompress the 
				files yet.  My biggest problem:  2-byte ints 
				and unsigneds in BC4, but 4-byte ints and
                              	unsigneds in C++Builder.
		12/25/97 RSB	Checked out using QREAD as the browser.  Works,
        			but I don't see how to avoid getting LOTS of
                              	QREAD windows open in practice, so I went back
                              	to my own browser.
		12/26/97 RSB	Got rid of some of the nastier bugs I was 
				seeing.  For future reference:
                              	1. The file-open dialog changes the current
                                   directory.  So if there are files there 
				   you're expecting to find later, you'd better 
				   save the directory name first thing when you 
				   start.
				2. Opening forms in a secondary thread doesn't
                              	   work very well, and results in a lot of queer
                                   and intermittent problems in the forms.
				Also, began the process of letting the program
                                decompress files.
		01/04/98 RSB	Stuff I've been working on, though not
               			necessarily completed yet:
                              	1. Converting my list-boxes of search results
                                   to tree views which are more useful.
				2. Beginnings of direct file browsing rather
                              	   than conversion to a text file followed by
                                   browsing the complete file.
				3. Global data encapsulated to local data and
                              	   function parameters, except for truly
                                   global data such as compression/decompression
                                   constant tables and filename-translation
                                   tables.
		01/10/02 RSB	Began removing all Win32 and GUI dependencies,
				so as to produce a simple library suitable
				for use with general 32-bit compilers,
				particularly GNU gcc, and added a GPL
				disclaimer.  Some things which NEED to be
				taken care of, but which I've not done yet:
				enforcing little-endian ordering in the 
				external files; enforcing 16-bit and 32-bit
				datatypes where necessary.		   

  The DOS version of RIP existed prior to the Windows 95 version, and the
  intention is to maintain complete compatibility between them at the
  file level.  Similarly for the gcc version.  Only file retrieval is dealt
  with here; compression remains a separate source file.  The comments below
  are general in nature, and have not been restricted to the retrieval
  functionality.  Furthermore, the comments refer to the original design,
  and don't necessarily apply in detail if some 16-bit 'x86 restrictions
  have been removed.
  
  Design Goals
  ------------

  This is a text-indexing/retrieval/browsing system with the design goal
  of easily maintaining a database of all available English-language
  public domain etext, using PC-type computers, into the early 21st
  century and (with minor modification) into the indefinite future.
  At this writing (12/20/97), the database being maintained is about 3 Gbyte
  and 9000 files using a 90 MHz Pentium, and the point is that similar
  performance levels are supposed to be maintained into the future.  The
  database will grow larger, but computer speeds and mass storage will
  increase also, and the latter should compensate (hopefully more than
  compensate) for the former.  At the same time, acceptable though
  degraded performance levels should be available using CD-ROMs, 486 or
  386 clase machines, etc.

  The point in a project such as this is to openly accept certain
  performance limitations up front in order to avoid building in
  functional limitations that will render the system inoperative when the
  database size increases past a certain point.  For example, while it may
  be faster to perform various sorting and searching operations in memory
  rather than from disk, if this means that when the available memory in
  the machine is exhausted that the database stops working, this is no
  bargain.  Therefore, at every point the system has been crafted to work
  with databases of any size, with the main limitation being only the
  size of the hard disk.  On the other hand, in the Windows 95 version,
  this is much less of a consideration since much more memory is guaranteed
  to be available (what with virtual memory and all), so some of the
  performance limits are reconsidered.  Two other limitations are inherent
  in the system:  There is approximately a 256M limit on the size of the
  individual text files; and there is a total limit of 65536 total files
  in a database.  For a text-library database, the typical pre-compression
  size of a book is 400K-500K.  Therefore, a database is limited in
  practical terms to about 33 Gbytes.  In practical terms, only the 64K
  limit on the number of files is likely to become a problem in the future;
  and, as explained in a moment, this only because of operating-system
  limitations and not because of problems with the RIP software itself.
  The filesize limitation will never be a problem, since the largest
  single text file I've ever seen (and I've seen a lot of them) is about
  10M; besides, the files can always be split into smaller parts.

  The 64K limit on number of files applies not to the full database, but only
  to any individual segment of it.  Because at this point it is necessary
  to use CD-ROM as the storage medium, there is a provision for dividing
  the database into segments so that they can be written to CD-ROMs, and so
  the CD-ROMs can work independently.  The segments aren't limited to 650K,
  and there's no practical limit on the number of segments (we actually
  artificially place a limit of 255, but we'll find below that this limit is
  not a problem), so with enough segments there's no limit to total
  number of files or total database size.  Unfortunately, search time
  increases linearly with the number of database
  segments and only logarithmically with the size of the segments, so it's
  still advantageous to make the segments as large as possible.

  However, there's a separate reason for the 64K-files limit, and
  working around this requires some thought:  Simply, any DOS/Windows
  disk can't have more than 64K files on because of operating system
  limitations.  But even this can be overcome.  Removable media types are
  becoming larger (650M CD-ROM is even now giving way to larger alternatives,
  which presumably will give way to even larger alternatives in the future),
  and EACH removable storage unit can hold 64K files.  Thus, when removable
  media reaches a 33G size, we can store an actual practical 64K files on
  each disk (or whatever).  Another alternative is partitioning of disks,
  each disk PARTITION can hold 64K files.  But in either case, there can
  only be so many segments:  just so many CD-ROMs you feel like swapping
  back and forth, or so many drive letters that can be used for partitions.
  For the sake of argument, we can suppose the maximum on each is 16 -- i.e.,
  16 disk partitions or 16 removable storage units.  Therefore, we can say
  that the practical limit on the total database size we can ever reach with
  DOS or Windows using RIP is about 1M files, for about 500G of text.

  I looked for shareware tools for this project prior to working on it
  myself, and discovered that while many such programs existed, they
  were all highly unsatisfactory in practice.  I found many defects in
  these programs, and had to work very hard to avoid falling into the
  same traps they did.  Significantly, these programs usually thought 10M
  would be a very large database; wherease I think 100G is a large database
  and 10M is a large FILE.  Usually a program would suffer from all the
  defects mentioned below, although some suffered only a few.  I hope I
  have avoided all of them.  Typical defects:

  1.  Very quick to add files when the database was small, but decreasing
      exponentially in speed as files new files were added.

  2.  Exhaustion of memory and eventual total failure as the database
      increased in size.

  3.  Failure of program or even system crash when indexing certain files.

  4.  No use of file compression.

  5.  Indexes of enormous size, so that the combined text and index might
      be twice the size of the original.

  6.  No integrated browser that I would care to use.  (I admit that my
      DOS-version browser isn't great, though it isn't bad, but surprisingly
      it grew on me and I actually enjoyed using it.  However, I hope the
      Windows 95 version will be a great improvement.)

  In fact, I never found a system that I would trust with even a 20M
  database, so in the end I needed to write my own.  Here are the basic
  engineering requirements for the task.  ("N" is the size of the
  database.)

  a.  Search time should be proportional to log N.

  b.  File preparation time (compression/indexing) should be proportional
      to N log N, and should be incremental so that it is not necessary to
      prepare the entire database in adding a new file.

  c.  Combined size of database (compressed files and index) should be
      some fixed proportion (but smaller) of the plain text.

  d.  RAM usage fixed, and not increasing with the size of the database.

  I have achieved these things.  On a 90 MHz Pentium, a 750MB/1500-file
  database can be totally prepared in 3-4 hours.  (I didn't time it.)
  A small number of files can be incrementally added in about 1/2 hour.
  The compressed database is about 75% of the size of the original.
  The program runs even on 512K machines, though 640K is helpful for
  file preparation.  Search time is a fraction of a second for negative
  single-word searches, though searches for phrases consisting of
  relatively common words can probably take up to a minute.  In the Windows
  95 version, I improve this a lot by making database searches a separate
  execution thread.

  SYSTEM OVERVIEW
  ---------------

  Text files are compressed/indexed using the "RIP C" command line option in
  the DOS version of the program.  I haven't seen any reason to add any of
  the compression or indexing stuff to the Windows 95 version as of yet.
  The text is compressed using the context-sensitive Huffman codes
  described below, and then indexing information is added to the file.
  The original text file is deleted.  The compressed file has the extension
  "rip".

  The text is processed in blocks of 4096 characters, although for
  indexing purposes the block size is extended somewhat so that the blocks
  overlap.  Pointer information concerning the blocks is stored so that the
  blocks can be individually accessed and decompressed (for viewing or
  searching) without the necessity of decompressing the entire file.  Each
  block is numbered with a 16-bit value.  Thus the 256M file limitation:
  4K*64K=256M.

  Indexing of a file occurs as follows.  First, "trash" words are discarded.
  These are supposed to be very common words (such as "and", "but", etc.)
  that occur so often it is better just to assume they are always present,
  rather than clogging up the index with them.  For execution speed, though,
  rather then discard the most common words we simply discard all words of
  3 characters or less.  (We keep all words containing Z, Q, J, or X as
  exceptions to this, so OZ or QUO would still be indexed.)  Next, each
  word is assigned a 32-bit hashcode, computed arithmetically from the
  ASCII representation.  Since there are only a few hundred thousand
  English words, and 4 billian 32-bit hashcodes, it is hoped that there
  is very little overlap and that the hashcodes are almost unique --
  one word, one hashcode.  Anyhow, a list is generated; for each unique
  hashcode appearing in the file is given the blocks in which the
  hashcode appears.  This list is in increasing order by hashcode, and
  secondarily in increasing order by block.  Finally, a "master index"
  is produced for the file giving simply all the hashcodes appearing in
  the file and the file offsets of their block lists.

  How would this index be used?  Suppose for example that you wanted to
  locate the word "Einstein".  You would compute the hashcode for
  "Einstein", and then examine the master index of the file.  The
  master index is a sorted list (by hashcode) of fixed-length records.
  A fast binary search would then determine instantly whether the
  word (or at least its almost-unique hashcode) appeared in the file.
  If the hashcode doesn't appear, the word is guaranteed not to appear.
  If the hashcode does appear, the word is probably present (with a very
  high degree of probability), but it can be determined fully as follows.
  Get the block list for the hashcode.  Decompress each of those blocks in
  turn and do a plain-text search.  Slower than a fully indexed system,
  of course, but still pretty fast.

  This is fast for an individual file, but slows down quite a lot if you
  have 1000+ files, particularly on CD-ROM, so there is an additional
  preparation step, "rip i".  This produces a master index file for the
  database as a whole.  The master index for the database is almost
  exactly like the master index for an individual file, except that instead
  of giving the block numbers in which each hashcode appears, it gives
  the file numbers.  (There is a separate file listing all the text files,
  and the file numbers are determined by this.)  So the system can determine
  quickly all the files in which a given hashcode appears, and from there
  can determine quickly all the blocks in each each of these file.

  When the 64K-file limit has to be extended, it can be done entirely at
  the level of the master index.  The individual text files do not need
  conversion.

  COMPRESSION
  -----------
  The technique is simple.  I have pre-analyzed some text files, more
  or less at random from what was my reading list at the momemt, and have
  generated an English-language Huffman code on the assumption that this is
  a reasonable average case.  The compression achieved by this means wasn't
  amazingly good (only about a 45% reduction), but I found that I could
  improve it a lot (>55-60% reduction) if I had context-varying (but not
  adaptive) Huffman codes.  In other words, a different Huffman code is used
  for the character following an "A" than for one following a "B".  I
  actually recognize 29 different contexts: A-Z, "digit", "punctuation",
  and "other".  ("Punctuation" is .,?!:;)  It is easy to see why this works
  better.  For example, "Q" is almost always followed by "U".  Punctuation
  is almost always followed by whitespace.  Digits are seldom followed by
  alphabetic characters.  And so on.  This means that if you know what
  went before, you can encode what follows in less bits, on the average.
  A really interesting example is "E".  This has the shortest Huffman code
  on the average, since it is the most common letter.  And yet it rarely
  (in comparison) appears at the beginnings of words.  It is common only
  in the interiors of words.  The Huffman tables are given by ASCII files
  (rip_a.dat, ..., rip_z.dat, ripdigit.dat, rippunct.dat, ripother.dat),
  so they can be replaced easily.  The *.dat files are generally produced
  by machine, for example by the huffman1.c program.  However, the
  program automatically uses them to create a smaller binary file,
  rip_allh.dat, to speed up program loading.  To replace the Huffman tables
  it is therefore necessary to erase rip_allh.dat, replace the other
  rip*.dat files, and then to run RIP.  NOTE that this will render any
  already-compressed files permenently unreadable, since they will be
  decompressed with different Huffman tables than they were compressed with.

  Exact compressed file format:
  -----------------------------

  Bytes 0-3		"RIP4"
  Bytes 4-7		Number of uncomressed bytes, LSB first.
  Bytes 8-11		Block size, LSB first.
  Bytes 12-15		Pointer to master index data.
  Bytes 16-		Table of longs (LSB first) giving bit offsets of the
			blocks.  There can be a maximum of 4G bits, so the
                    	max filesize is 536M bytes.  However, files of this
                     	size can't be achieved with the block size normally used
                     	since there is also a limit of 64K blocks in the file.
  ... compression data ...
			(A bit stream of Huffman codes)
  ... index data ...
			(For each hashcode actually appearing in the file,
			a list of unsigneds, in increasing order, giving the
			block numbers in which the hashcode appears.  No
			counts or terminators are used.)
  ... master index data ...
			(First, long unsigned integer giving the total number
			of distinct hashcodes appearing in the file.  This
			includes a terminator hashcode of 0, which is otherwise
			an unused value.  This is followed by a table of
			records, with each record being a pair of longs.  The
			first long is a hashcode, and the second is the
			offset of this hashcode in the index data.  The
			hashcodes appear in increasing order both in the index
			data and in the master index data.  There is a
			Terminating record with hashcode 0 and offset pointing
			to the master index data itself.  This arrangement
			allows each calculation for each hashcode of how many
			blocks are used.)

  In other words, to find a given word, you do a binary search of the
  "master index data" to see if the word's hashcode is present.  If not,
  the word isn't there.  If the word is present, the "master index data"
  table will tell you where to look in the "index data" to get a list of
  all blocks in which the word (or at least its hashcode) appears.  If
  your'e worried about it, you can then uncompress just those specific
  blocks to find out if the word ACTUALLY appears (rather than just its
  hashcode).

  The total number of blocks isn't given, since it is the number of
  uncompressed bytes, divided by the blocksize and rounded up.

  rip.cfg
  -------

  The rip.cfg file used in the DOS version is irrelevant to the Windows 95
  and gcc versions.

  rip.lst
  -------

  Generating a single huge index file for the entire database is 
  very inefficient for several reasons.  For one thing, when we 
  surpass the 64K-file limit, the system will have to be
  re-written.  Second, since as the database grows and we make a new
  CD each time we pass a 650K limit, the databases for these CDs need only
  be generated once.  A more efficient scheme than generating a single
  huge index is to allow instead a master list of indices:  on this list we
  would have one entry for each finalized CD, and probably one for work in
  progress.  Therefore, we never have to generate an index larger than that
  needed for work in progress.  Also, this effectively increases the number
  of files allowed to 64K*max_indexes.

  Anyhow, that's what rip.lst is:  a master list, with one line for each
  searchable index.  In its simplest form, each record simply is the base
  name of the index.  For example, if we copied all the indices (*.fil and
  *.ind) into a single directory and called them vol1.fil/vol1.ind,
  vol2.fil/vol2.ind, ..., rip.fil/rip.ind (the latter being for work in
  progress and the former for finalized CDs), we could simply have a
  rip.lst like this:

	rip
	vol1
	vol2
	.
	.
	.

  The databases would be searched in the order shown.

  One slight difficulty with this is the amount of disk space required by
  all these indices (30M+ for a typical CD index), so we might want to keep
  the CD indices on the CDs themselves.  If we do this, Some earlier
  versions of RIP used a master index named rip.fil/rip2.ind (rather than
  rip.fil/rip.ind).  To add the "2" to the IND-file name, we put a second
  field of "*" in the record:

	rip
	/cdrom/rip *

  Thus, the working database segment (that is, what will become the next
  CD, but which isn't big enough yet for that) has its index files
  (rip.fil/rip.ind) on the hard disk, while the completed CD segment has
  its index files (rip.fil/rip2.ind) on the CD.

  Another difficulty is that the leading portions of paths in CD-based rip.fil
  files are generally wrong, since the CDs are moved from computer to computer.
  Furthermore, it has even happened that the base directory 
  (as well as the drive designation) has been wrong on the
  CD, because the index was originally generated with the files on the
  hard disk, and then merely copied to the CD.  For example, the pathnames
  in the CD rip.fil might be D:\RIP\something (having originated on a
  Windows computer), when what is really wanted is /cdrom/something.  
  The "something" is correct, but the leading part of the
  pathname is wrong.  This can be handled by adding two more fields to
  the rip.lst records:

		rip
		/cdrom/rip * D:\RIP\ /cdrom/

  This would simply replace D:\RIP\ by /cdrom/ in the leading positions of the
  filenames found in /cdrom/rip.fil, but would let the pathnames in rip.fil
  unchanged.

  MASTER TOC
  ----------

  Helpful Hint:  In really large databases, when you are simply trying to
  find a specific book by title or author, a free-form search of the
  entire database is a hassle.  It would be more convenient to just search
  a title/author list.  Of course, there is no such thing.  However, the
  "rip n" command can be used to generate a master list of the beginnings
  (say, the first 10 lines) of all the files, and if this master list itself
  is treated just like any old text file (i.e., compressed and indexed), and
  is the very first file in the first index, then the system will be able
  to complete title/author searches much quicker, since we are guaranteed
  that most titles and authors are in the first first file searched.  This
  allows the user to determine the desired filename, and from there to
  browse the file directly using its filename rather than a database-search
  operation.

*/

#include <stdlib.h>
#include <stdio.h>
/* #include <ctype.h>  Commented out because isdigit() and isalpha()
   didn't work right for me when values outside of 32..126 were used.
   I had to fake them up with comparisons!  */
#include <string.h>
#if 0
#include <process.h>
#include <dir.h>
#include <io.h>
#include "Rip97.h"
#include "error.h"
#endif

#if 0
/* Changes cursor shape */
void
CursorShape (TCursor shape)
{
  Main->Cursor = shape;
  Main->FileList->Cursor = shape;
  Main->Memo->Cursor = shape;
  Main->RawTree->Cursor = shape;
  Main->GradedTree->Cursor = shape;
  Main->MatchTree->Cursor = shape;
}
#endif

/* Miscellaneous dummy strings. */
static char s[1000];
static char s1[256], s2[256], s3[256], s4[256];

typedef long HASHVALUE;
#define MAX_HASHTHINGS 200
static HASHVALUE hashthings[MAX_HASHTHINGS];

/*
  The stuff below is used to Translate the names of the index files and
  drives on which the compressed data files are stored.
*/
#define MAX_INDICES 100
#define MAX_INDEXNAME 21
#define MAX_INDEXTRAN 11
typedef struct
{
  char star2;
  char name[MAX_INDEXNAME];
  char src[MAX_INDEXTRAN];
  char dst[MAX_INDEXTRAN];
}
INDEXFILE;
static INDEXFILE indices[MAX_INDICES];
static unsigned short int num_indices = 0;

char HomePath[257];
void
initialize_rip (void)
{
  char *ss;
  int i;

  getcwd (HomePath, sizeof (HomePath));
  Main->RawTree->Items->Clear ();
  Main->GradedTree->Items->Clear ();

  /*
     Initialize all the static data if it hasn't been done before.  Includes:
     Constants used for hashcode generation.
     In-memory version of rip.lst.
   */
  FILE *files;
  {
    HASHVALUE hi;
    unsigned short int i;
    for (hi = 1, i = 0; i < MAX_HASHTHINGS; i++, hi *= 29)
      hashthings[i] = hi;
  }
  sprintf (s, "%s\\RIP.LST", HomePath);
  files = fopen (s, "r");
  if (files == NULL)
    {
      strcpy (indices[0].name, "RIP");
      num_indices = 1;
    }
  else
    {
      while (NULL != fgets (s, sizeof (s), files))
	{
	  for (ss = s; *ss; ss++)
	    if (*ss == '\n')
	      *ss = 0;
	  *s1 = *s2 = *s3 = *s4 = 0;
	  i = sscanf (s, "%s %s %s %s", s1, s2, s3, s4);
	  if (!i)
	    continue;
	  s1[MAX_INDEXNAME - 1] = 0;
	  strcpy (indices[num_indices].name, s1);
	  if (i > 1)
	    {
	      if (*s2 == '*')
		{
		  indices[num_indices].star2 = 1;
		  strcpy (s2, s3);
		  strcpy (s3, s4);
		  i--;
		}
	      if (i > 1)
		{
		  s2[MAX_INDEXTRAN - 1] = 0;
		  s3[MAX_INDEXTRAN - 1] = 0;
		  strupr (s2);
		  strupr (s3);
		  strcpy (indices[num_indices].src, s2);
		  strcpy (indices[num_indices].dst, s3);
		}
	    }
	  num_indices++;
	  if (num_indices >= MAX_INDICES)
	    break;
	}
      fclose (files);
    }
  {
    int ReadHuffmanCodes (void);
    if (ReadHuffmanCodes ())
      Main->QueryDatabase->Visible = False;
  }
}

/* Sets up file-browsing. */
void
BrowseFile (char *name, int offset, char *DisplayedName)
{
  CursorShape (crHourGlass);
  Main->Memo->Visible = False;
  if (offset < 0)
    offset = 0;
  Main->Memo->Lines->LoadFromFile (name);
  /* JustifyText(buf); */
  Main->Memo->SelStart = offset;
  Main->Memo->SelLength = 0;
  Main->Close->Enabled = True;
  Main->Print->Enabled = True;
  Main->Edit->Enabled = True;
  Main->FindWord->Enabled = True;
  Main->SearchAgain->Enabled = True;
  Main->Memo->Visible = True;
  Main->Memo->SetFocus ();
  Main->StatusBar->Panels->Items[0]->Text = DisplayedName;
  CursorShape (crDefault);
}

/* Check if a word is a trash (unindexed) word or not */
#define MINSTRING 4		/* min size of non-trash words */
int
iswimp (char *s)
{
  int j;
  if (strlen (s) >= MINSTRING)
    return (0);
  for (; *s; s++)
    {
      j = toupper (*s);
      if (j == 'Z' || j == 'J' || j == 'Q' || j == 'X')
	return (0);
    }
  return (1);
}

HASHVALUE
hash (char *s)
{
  HASHVALUE hi;
  short int i;
  for (i = 0, hi = 0; *s; s++)
    {
      if (isalpha (*s))
	{
	  hi += hashthings[i] * (toupper (*s) - '@');
	  i++;
	}
    }
  if (i && !hi)
    return (-1);
  else
    return (hi);
}


typedef struct
{
  HASHVALUE hash;
  unsigned long offset;
}
NDX;

#define SEARCHSTRING_SIZE 129
typedef struct
{
  HASHVALUE hash;		/* =0 for wimps */
  unsigned long start_offset, end_offset;
  char string[SEARCHSTRING_SIZE];
}
SEARCH;
#define MAX_SEARCH 16
static SEARCH search[MAX_SEARCH];
static unsigned short int num_search;


typedef struct
{
  unsigned short int next_search, max_search;
  char file_block[0x2000];
}
HASHBITS;
unsigned short int
nexthash (HASHBITS * hashbits)
{
  unsigned int byte, mask;
  for (; hashbits->next_search <= hashbits->max_search;
       hashbits->next_search++)
    {
      byte = hashbits->next_search >> 3;
      mask = 1 << (hashbits->next_search & 7);
      if (hashbits->file_block[byte] & mask)
	break;
    }
  if (hashbits->next_search <= hashbits->max_search)
    return (hashbits->next_search++);
  return (0xffffU);
}

unsigned short int
findhash (FILE * f, HASHBITS * hashbits)
{
  HASHVALUE num_hashes, mx;
  unsigned short int block;
  unsigned short int i, j, found;
  unsigned int byte, mask;
  unsigned long n;
  unsigned long index_start;
  HASHVALUE mn, m;
  NDX current_ndx;

  block = 0xffffU;
  hashbits->max_search = hashbits->next_search = 0;
  for (i = 0; i < 0x2000; i++)
    hashbits->file_block[i] = 0xff;

  fread (&num_hashes, sizeof (num_hashes), 1, f);
  if (num_hashes == 0 || num_hashes == -1)
    goto done;

  index_start = ftell (f);
  for (found = 2; found < num_search; found++)
    if (search[found].hash)
      {
	mn = 0;
	mx = num_hashes;
	for (; mx > mn;)
	  {
	    m = (mx + mn) / 2;
	    fseek (f, index_start + sizeof (NDX) * m, SEEK_SET);
	    fread (&current_ndx, sizeof (NDX), 1, f);
	    if (search[found].hash < current_ndx.hash)
	      mx = m;
	    else if (search[found].hash > current_ndx.hash)
	      mn = m + 1;
	    else
	      {
		search[found].start_offset = current_ndx.offset;
		fread (&current_ndx, sizeof (NDX), 1, f);
		search[found].end_offset = current_ndx.offset;
		break;
	      }
	  }
	if (mx <= mn)
	  break;
      }
    else
      found++;

  /* What we do now is to set up the file_block array so that every
     block found is in it. */
  if (found >= num_search)
    {
      for (i = 2; i < num_search; i++)
	{
	  block = 0;
	  if (search[i].hash)
	    {
	      n = search[i].start_offset;
	      fseek (f, n, SEEK_SET);
	      while (n < search[i].end_offset)
		{
		  fread (&j, sizeof (j), 1, f);
		  /* printf("Marking block %u for parm %u\n",j,i); */
		  if (j > hashbits->max_search)
		    hashbits->max_search = j;
		  for (; block < j; block++)
		    {
		      byte = block >> 3;
		      mask = 1 << (block & 7);
		      hashbits->file_block[byte] &= ~mask;
		    }
		  block = j + 1;
		  n += 2;
		}
	      for (; block < 0xffffU; block++)
		{
		  byte = block >> 3;
		  mask = 1 << (block & 7);
		  hashbits->file_block[byte] &= ~mask;
		}
	    }
	}
      block = nexthash (hashbits);
    }

done:
  return (block);
}

void
filename_translate (char *filename, int xlation_index)
{
  int n;
  if (xlation_index >= num_indices)
    return;
  strupr (filename);
  n = strlen (indices[xlation_index].src);
  if (!n)
    return;
  if (!strncmp (filename, indices[xlation_index].src, n))
    {
      strcpy (s1, filename + n);
      sprintf (filename, "%s%s", indices[xlation_index].dst, s1);
    }
}

/* Terminates a string at a newline. */
#define AUTOTRIM
#ifdef AUTOTRIM
void
trim (char *s)
{
  for (; *s; s++)
    if (*s == '\n' || *s == '\r')
      {
	*s = 0;
	break;
      }
}
#endif

/* Reads the header of a compressed file.  In the DOS version, the various
   HEADER variables were globals.  However, this means the code wouldn't be
   reentrant, leaving me with a problem if the browser thread and the
   search thread both try to use the same stuff. */
typedef struct
{
  FILE *fin;
  unsigned long num_text, blocksize, index, blocks;
  char g_lastchar;
  unsigned int g_kb, g_bits, g_data;
  int lastchar, ignore_cr;
}
HEADER;
short int
read_header (HEADER * header)
{
  if (getc (header->fin) != 'R' ||
      getc (header->fin) != 'I' ||
      getc (header->fin) != 'P' || getc (header->fin) != '3')
    return (1);
  if (1 != fread (&header->num_text, 4, 1, header->fin))
    return (2);
  if (1 != fread (&header->blocksize, 4, 1, header->fin))
    return (3);
  header->blocks =
    (header->num_text + header->blocksize - 1) / header->blocksize;
  if (1 != fread (&header->index, 4, 1, header->fin))
    return (4);
  header->lastchar = ' ';
  header->g_lastchar = 0;
  header->g_kb = 0;
  header->g_bits = 0;
  return (0);
}

/* create an output file */
FILE *
create_file (char *fil, int overwrite)
{
  FILE *fout;
  fout = fopen (fil, "r");
  if (fout != NULL)
    {
      fclose (fout);
      if (!overwrite)
	return (NULL);
    }
  fout = fopen (fil, "w+b");
  if (fout == NULL)
    return (NULL);
  setvbuf (fout, NULL, _IOFBF, 8192);
  return (fout);
}

/* open the input file */
FILE *
open_file (char *fil)
{
  FILE *fin;
  fin = fopen (fil, "rb");
  if (fin == NULL)
    return (NULL);
  setvbuf (fin, NULL, _IOFBF, 16384);
  return (fin);
}

/*
  This is the function which is called to service a request to search the
  RIP database.  It is called as a new thread, allowing the user to continue
  to examine the text window in the browser while it operates, and so it has
  the leisure to do a lot more stuff than the DOS version could.  All it
  really does is to add filenames and offsets to a list box, and in order to
  get quick responses, it proceeds in phases, with some quick steps being
  performed at the beginning of the search to identify candidate files, and
  with progressively more time consuming but more accurate steps following
  later to weed out many of the candidates and/or to grade the survivors by
  relevance.

  Of course, you don't want the search to continue indefinitely, so there is
  a global flag you can use to tell the function to shut down the thread.
  The flag shutdown_request can be used to request a shutdown, while the
  flag shutdown_status is used to report whether a shutdown has occurred.
  Normally, shutdown_status is 1, but it is set to 0 when the search thread
  begins executing; it goes back to 1 after the thread has completed.
*/

volatile int shutdown_status = 1;
volatile int shutdown_request = 0;

void
rip_query (void *arglist)
{
  int i, j, err;
  unsigned short int indexno, fileno;
  char outfile[257];
  unsigned int num_refined;

  shutdown_status = shutdown_request = 0;

  Main->QueryDialog->CloseDialog ();
  num_refined = 0;
  Main->RawTree->Items->Clear ();
  /* Main->RawTree->FullExpand(); */
  Main->GradedTree->Items->Clear ();
  /* Main->GradedTree->FullExpand(); */
  Main->MatchTree->ActivePage = Main->Raw;

  err = sscanf (Main->QueryDialog->FindText.c_str (),
		"%s %s %s %s %s %s %s %s %s %s %s %s %s %s",
		search[2].string, search[3].string, search[4].string,
		search[5].string, search[6].string, search[7].string,
		search[8].string, search[9].string, search[10].string,
		search[11].string, search[12].string, search[13].string,
		search[14].string, search[15].string);
  if (err == EOF || err == 0)
    goto exit;
#pragma warn -sig
  num_search = err + 2;
#pragma warn +sig
#ifdef AUTOTRIM
  for (i = 2; i < num_search; i++)
    trim (search[i].string);
#endif

  /* Eliminate wimp words and get hashcodes of non-wimps. */
  for (i = 2; i < num_search; i++)
    {
      if (iswimp (search[i].string))
	{
	  num_search--;
	  for (j = i; j < num_search; j++)
	    search[j] = search[j + 1];
	  i--;
	}
      else
	{
	  search[i].hash = hash (search[i].string);
	}
      search[i].start_offset = 0;
    }

  /*
     First and quickest search phase:  find all the files that contain
     the hashcodes.  It's fast because the master index files (which are
     all that's needed) are relatively small and usually on hard disk.
     It's inaccurate, though; for example, if the word "albert" is at the
     end of a 1M file, and the word "einstein" is at the front of it, the
     file will still match "albert einstein".
   */
  /* Loop on the index files */
  for (indexno = 0; indexno < num_indices; indexno++)
    {
      FILE *indexfile, *files;
      char infile[257];
      unsigned long n;
      HASHBITS hashbits;
      if (shutdown_request)
	goto exit;
      if (indices[indexno].star2)
	sprintf (outfile, "%s2.IND", indices[indexno].name);
      else
	sprintf (outfile, "%s.IND", indices[indexno].name);
      indexfile = fopen (outfile, "rb");
      if (indexfile == NULL)
	continue;
      fseek (indexfile, 0, SEEK_SET);
      fread (&n, sizeof (n), 1, indexfile);
      fseek (indexfile, n, SEEK_SET);
      sprintf (outfile, "%s.FIL", indices[indexno].name);
      files = fopen (outfile, "r");
      if (files == NULL)
	{
	  fclose (indexfile);
	  continue;
	}
      for (fileno = findhash (indexfile, &hashbits);
	   fileno != 0xffffU; fileno = nexthash (&hashbits))
	{
	  HEADER header;
	  unsigned short int block;
	  HASHBITS hashbits;
	  if (shutdown_request)
	    {
	      fclose (indexfile);
	      fclose (files);
	      goto exit;
	    }
	  fseek (files, 62L * fileno, SEEK_SET);
	  fscanf (files, "%s", infile);
	  filename_translate (infile, indexno);
	  /* At this point, we know a filename, but we need to identify the
	     specific blocks of the candidate file. */
	  header.fin = open_file (infile);
	  if (header.fin == NULL)
	    continue;
	  if (read_header (&header))
	    {
	      fclose (header.fin);
	      continue;
	    }
	  fseek (header.fin, header.index, SEEK_SET);
	  for (block = findhash (header.fin, &hashbits);
	       block != 0xffffU; block = nexthash (&hashbits))
	    {
	      sprintf (outfile, "%s %d", infile, block * header.blocksize);
	      Main->RawTree->Items->Add (outfile);
	      num_refined++;
	      sprintf (outfile, "Raw: %d", num_refined);
	      Main->StatusBar->Panels->Items[1]->Text = outfile;
	    }
	  fclose (header.fin);
	  /* if (j) parent->Expand(0); */
	}

      fclose (indexfile);
      fclose (files);
    }

  /*
    Second, more detailed search phase.  Not certain yet what it does, but
    probably reads back actual file blocks and does some kind of string
    operations on them.
   */


exit:
  sprintf (outfile, "Done: %d", num_refined);
  Main->StatusBar->Panels->Items[1]->Text = outfile;
  shutdown_status = 1;
  /* _endthread(); Not needed. */
}

/*
  The Huffman code tables used for compression and for decompression are
  in different formats.  For compression, for each possible character
  (0 to 255) we simply need to know how many bits are used, and what the
  bits are.  There are always less than 32 bits, so we can use a long
  to store the bits.  For decompression, since a variable number of bits
  is used, we need to make a binary tree.
*/
typedef struct
{
  unsigned short int numbits;
  unsigned long bits;
}
COMPRESS;
typedef COMPRESS COMPRESSOR[256];

/* 
  "zero" is the offset in the array storing the tree of the zero branch.
  if zero==0, then there is none.  Similarly, "one" is the offset of the
  one branch, and one=0 if there isn't one.  If both are zero, then
  "value" is the decoded byte.  Otherwise, "value" is undefined. 
*/
typedef struct
{
  char value;
  unsigned short int zero;
  unsigned short int one;
}
DECOMPRESS;
/* 
  MAXD is the maximum number of nodes in the tree.  It might seem as though
  256 is enough, but since each atomic node must have a non-atomic node
  pointing to it, 256 cannot be enough.  On the other hand, each non-atomic
  node must actually point to two other nodes (atomic or non-atomic), rather
  than just one, since otherwise the tree wouldn't be minimal.  In fact
  it is easy to see by induction that any minimal tree with N atomic nodes
  must have N-1 non-atomic nodes. 
*/
#define MAXD 511
typedef DECOMPRESS DECOMPRESSOR[MAXD];

/* 
  In the Windows version, unlike the DOS version, we can simply set up these
  things as arrays, rather than allocate them at runtime.  Couldn't do that
  in DOS because the data size exceeds 64K. 
*/
DECOMPRESSOR decompressors[29];
COMPRESSOR compressors[29];

/* Read in a single-context code table.  The table is in the format
   output by huffman1.c. */
short int
read_huffman (char *filename, int c_index)
{
  char s[257], ss[257], *sss;
  short int i, k;
  int j;
  unsigned long n;
  unsigned short int node, numd;
  COMPRESSOR *compress;
  DECOMPRESSOR *decompress;
  FILE *fin;

  fin = fopen (filename, "r");
  if (fin == NULL)
    return (1);
  for (; NULL != fgets (s, sizeof (s), fin);)
    {
      if (!strncmpi (s, "CODES:", 6))
	goto codes;
    }

huffman_error:
  fclose (fin);
  return (2);

  /* load the code table */
codes:
  compress = &compressors[c_index];
  decompress = &decompressors[c_index];
  (*decompress)[0].one = 0;
  (*decompress)[0].zero = 0;
  numd = 1;
  for (i = 0; i < 256; i++)
    {
      if (NULL == fgets (s, sizeof (s), fin))
	goto huffman_error;
      if (2 != sscanf (s, "%u:%s", &j, ss))
	goto huffman_error;
      if (j > 255 || j < 0)
	goto huffman_error;
      n = 0;
      k = 0;
      node = 0;
      for (sss = ss; *sss; sss++)
	{
	  n += n;
	  if (*sss == '1')
	    {
	      n++;
	      if ((*decompress)[node].one == 0)
		{
		  if (numd >= MAXD)
		    goto huffman_error;
		  (*decompress)[node].one = numd;
		  (*decompress)[numd].one = 0;
		  (*decompress)[numd].zero = 0;
		  numd++;
		}
	      node = (*decompress)[node].one;
	    }
	  else if (*sss == '0')
	    {
	      if ((*decompress)[node].zero == 0)
		{
		  if (numd >= MAXD)
		    goto huffman_error;
		  (*decompress)[node].zero = numd;
		  (*decompress)[numd].one = 0;
		  (*decompress)[numd].zero = 0;
		  numd++;
		}
	      node = (*decompress)[node].zero;
	    }
	  else
	    goto huffman_error;
	  k++;
	}
      (*decompress)[node].value = j;
      if (k > 32)
	goto huffman_error;
      (*compress)[j].numbits = k;
      (*compress)[j].bits = n;
    }
  fclose (fin);
  return (0);
}

/* Read in the complete set of Huffman codes for later use. */
int
ReadHuffmanCodes (void)
{
  FILE *fin;
  char infile[257];
  int i;
  sprintf (infile, "%s\\rip97.dat", HomePath);
  fin = fopen (infile, "rb");
  if (fin == NULL)
    {
      fin = fopen (infile, "wb");
      for (i = 0; i < 29; i++)
	{
	  if (i < 26)
	    sprintf (infile, "%s\\RIP_%c.DAT", HomePath, 'A' + i);
	  else if (i == 26)
	    sprintf (infile, "%s\\RIPDIGIT.DAT", HomePath);
	  else if (i == 27)
	    sprintf (infile, "%s\\RIPPUNCT.DAT", HomePath);
	  else if (i == 28)
	    sprintf (infile, "%s\\RIPOTHER.DAT", HomePath);
	  if (read_huffman (infile, i))
	    {
	      fclose (fin);
	      return (1);
	    }
	  fwrite (compressors[i], sizeof (COMPRESSOR), 1, fin);
	  fwrite (decompressors[i], sizeof (DECOMPRESSOR), 1, fin);
	}
    }
  else
    {
      for (i = 0; i < 29; i++)
	{
	  fread (compressors[i], sizeof (COMPRESSOR), 1, fin);
	  fread (decompressors[i], sizeof (DECOMPRESSOR), 1, fin);
	}
    }
  fclose (fin);
  return (0);
}

short int
ispunctuation (int c)
{
  if (c == '.')
    return (1);
  if (c == '!')
    return (1);
  if (c == '?')
    return (1);
  if (c == ':')
    return (1);
  if (c == ';')
    return (1);
  if (c == ',')
    return (1);
  if (c == ')')
    return (1);
  return (0);
}

/* Expand a rip-file back to original form. */
#define BLOCKSIZE 4096L
#define OVERBLOCK (BLOCKSIZE+256)	/* block size for indexing */
int
decompress_it (char *infile, char *outf, int overwrite)
{
  char outfile[257];
  FILE *fout;
  int err, j;
  unsigned long n;
  unsigned int i, k, node, kb;
  struct ftime ft;
  char *sss;
  DECOMPRESSOR *decompress;
  HEADER header;

  kb = 0;
  for (sss = infile; *sss && *sss != '.'; sss++);
  if (strncmpi (sss, ".RIP", 4))
    return (3);

  if (strlen (outf) > 250)
    return (4);
  strcpy (outfile, outf);
  for (sss = outfile; *sss && *sss != '.'; sss++);
  strcpy (sss, ".txt");

  header.fin = open_file (infile);
  if (header.fin == NULL)
    return (0x10);
  fout = create_file (outfile, overwrite);
  if (fout == NULL)
    {
      fclose (header.fin);
      return (0x20);
    }
  err = read_header (&header);
  if (err)
    {
      err += 0x30;
    ErrorExit:
      fclose (header.fin);
      fclose (fout);
      return (err);
    }
  n = header.num_text;

  if (fseek (header.fin, 4 * header.blocks, SEEK_CUR))
    {
      err = 1;
      goto ErrorExit;
    }

  i = 0;			/* packed input bits */
  k = 0;			/* bit count */
  for (; n; n--)
    {
      for (node = 0;;)
	{
	  if (!k)
	    {
	      i = getc (header.fin);
	      if (((int) i) == EOF)
		{
		  err = 2;
		  goto ErrorExit;
		}
	      k = 8;
	    }
	  j = toupper (header.lastchar);
	  if (j >= 'A' && j <= 'Z')
	    {
	      decompress = &decompressors[j - 'A'];
	    }
	  else if (j >= '0' && j <= '9')
	    {
	      decompress = &decompressors[26];
	    }
	  else if (ispunctuation (j))
	    {
	      decompress = &decompressors[27];
	    }
	  else
	    {
	      decompress = &decompressors[28];
	    }
	  if (i & 128)
	    node = (*decompress)[node].one;
	  else
	    node = (*decompress)[node].zero;
	  if (node > 511)
	    exit (EXIT_FAILURE);
	  i = i << 1;
	  k--;
	  if (!(*decompress)[node].one && !(*decompress)[node].zero)
	    {
	      fputc (header.lastchar = (*decompress)[node].value, fout);
	      kb++;
	      if (kb == BLOCKSIZE)
		{
		  kb = 0;
		  header.lastchar = ' ';
		}
	      break;
	    }
	}
    }
  getftime (fileno (header.fin), &ft);
  fflush (fout);
  setftime (fileno (fout), &ft);
  if (ferror (fout))
    {
      err = 5;
      goto ErrorExit;
    }
  else
    {
      fclose (fout);
      fclose (header.fin);
    }
  return (0);
}

/* Browses a file without regard to filetype, processing it in advance
   if needed. */
void
Browse (char *FileName, int offset)
{
  char *ss;
  char outfile[257];
  if (!strcmpi (FileName, Main->StatusBar->Panels->Items[0]->Text.c_str ()))
    {
      Main->Memo->SelStart = offset;
      Main->Memo->SelLength = 0;
      return;
    }
  for (ss = FileName; *ss; ss++)
    if (*ss == '.')
      break;
  if (!strncmpi (ss, ".rip", 4))
    {
      CursorShape (crHourGlass);
      sprintf (outfile, "%s\\temp.txt", HomePath);
      decompress_it (FileName, outfile, 1);
      BrowseFile (outfile, offset, FileName);
    }
  else
    BrowseFile (FileName, offset, FileName);
}

/* Uncompress next byte in a file. The g_xxx variables are assumed to have
   been set up properly in advance, possibly by get_text().  At the
   beginning of a file, that appropriate values are the ones shown in
   the initializers.  Returns -1 on error. */
int
get_onebyte (HEADER * header)
{
  int c;
  unsigned int j, node;
  DECOMPRESSOR *decompress;

  for (node = 0;;)
    {
      if (!header->g_bits)
	{
	  header->g_data = getc (header->fin);
	  if (((int) header->g_data) == EOF)
	    return (-1);
	  header->g_bits = 8;
	}
      c = toupper (header->g_lastchar);
      if (c >= 'A' && c <= 'Z')
	{
	  decompress = &decompressors[c - 'A'];
	}
      else if (header->g_lastchar >= '0' && header->g_lastchar <= '9')
	{
	  decompress = &decompressors[26];
	}
      else if (ispunctuation (header->g_lastchar))
	{
	  decompress = &decompressors[27];
	}
      else
	{
	  decompress = &decompressors[28];
	}
      if (header->g_data & 128)
	node = (*decompress)[node].one;
      else
	node = (*decompress)[node].zero;
      header->g_data = header->g_data << 1;
      header->g_bits--;
      if (!(*decompress)[node].one && !(*decompress)[node].zero)
	{
	  j = header->g_lastchar = (*decompress)[node].value;
	  header->g_kb++;
	  if (header->g_kb == BLOCKSIZE)
	    {
	      header->g_kb = 0;
	      header->g_lastchar = ' ';
	    }
	  header->ignore_cr = (header->lastchar == '\r');
	  header->lastchar = j;
	  return (j);
	}
    }
}

/* Uncompress text beginning at a given address in the uncompressed file.
   "num_text" is the number of bytes in the file, if uncompressed.
   "num" is the desired number of bytes, while the actual number uncompressed
   is returned.  In general, the number uncompressed is always equal to the
   number requested unless the end of file is reached.  */
unsigned int
get_text (HEADER * header,
	  unsigned long loffset, char *text_buf, unsigned int num)
{
  unsigned long n;
  unsigned int block, offset, j;
  int i;

  header->lastchar = 0;
  header->ignore_cr = 0;

  block = loffset / header->blocksize;
  offset = loffset % header->blocksize;

  fseek (header->fin, 16 + 4 * block, SEEK_SET);
  fread (&n, 4, 1, header->fin);
  header->g_lastchar = ' ';
  header->g_kb = 0;
  header->g_bits = (n & 7);
  n /= 8;
  fseek (header->fin, n, SEEK_SET);
  header->g_data = fgetc (header->fin) << header->g_bits;
  header->g_bits = 8 - header->g_bits;
  if (loffset >= header->num_text)
    {
      num = 0;
      goto done;
    }
  if (loffset + num > header->num_text)
    num = header->num_text - loffset;
  for (; offset; offset--)
    if (-1 == get_onebyte (header))
      {
	num = 0;
	goto done;
      }
  for (j = 0; j < num; j++)
    {
      i = get_onebyte (header);
      if (i == -1)
	{
	  num = j;
	  goto done;
	}
      text_buf[j] = i;
    }
done:
  return (num);
}

/* Called when the list box is clicked. */
void
decompress (int BoxLevel)
{
  char text_buf[OVERBLOCK + 1];
  int i, j;
  int offset;
  char FileName[257], s[257];
  HEADER header;
  offset = 0;
  if (BoxLevel == 1)
    {
      i = Main->GradedTree->ItemIndex;
      strcpy (s, Main->GradedTree->Items->Strings[i].c_str ());
    }
  else if (BoxLevel == 0)
    {
      i = Main->RawTree->ItemIndex;
      strcpy (s, Main->RawTree->Items->Strings[i].c_str ());
    }
  else
    return;
  j = sscanf (s, "%s %d", FileName, &offset);
  if (j == 1)
    {
      Browse (FileName, offset);
    }
  else if (j == 2)
    {
      header.fin = open_file (FileName);
      if (header.fin == NULL)
	return;
      if (read_header (&header))
	{
	  fclose (header.fin);
	  return;
	}
      CursorShape (crHourGlass);
      Main->Memo->Visible = False;
      i = get_text (&header, offset, text_buf, OVERBLOCK);
      fclose (header.fin);
      text_buf[i] = 0;
      /* JustifyText(buf); */
      Main->Memo->SelStart = 0;
      Main->Memo->SelLength = 0;
      Main->Memo->Text = text_buf;
      Main->Close->Enabled = True;
      Main->Print->Enabled = True;
      Main->Edit->Enabled = True;
      Main->FindWord->Enabled = True;
      Main->SearchAgain->Enabled = True;
      Main->Memo->Visible = True;
      Main->Memo->SetFocus ();
      Main->StatusBar->Panels->Items[0]->Text = s;
      CursorShape (crDefault);
    }
}
